from weakref import WeakValueDictionary
import zipfile
from StringIO import StringIO

import pygame
from OpenGL.GL import *
from OpenGL.GLU import *

from qgl.scene import Color, Texture, Quad, Sequence, Text, Polyline
import qgl.texture


#cache for textures and fonts, so that they can be shared among nodes in the tree
qgl_cache = WeakValueDictionary()


class Visitor(object):
    def push_state(self, node): pass
    def pop_state(self): pass
    def visit_Node(self, node): pass
    def visit_Root(self, node): pass
    def visit_Viewport(self, node): pass
    def visit_Transform(self, node): pass
    def visit_Group(self, node): pass
    def visit_Switch(self, node): pass
        
        
class GLTexture(object):
    @classmethod
    def from_file(cls, filename, mipmap, filter):
        img = pygame.image.load(filename).convert(32, pygame.SRCALPHA)
        img = img.convert_alpha(img)
        return cls(img, mipmap, filter, flip=True)

    def __init__(self, img, mipmap, filter, flip=False):
        img = img.convert(32, pygame.SRCALPHA)
        img = img.convert_alpha(img)
        
        if not mipmap:
            assert img.get_size()[0] in [int(2**i) for i in xrange(100)], "width of image not a power of two."
            assert img.get_size()[1] in [int(2**i) for i in xrange(100)], "height of image not a power of two."

        ptype = GL_RGBA
        channels = 4
        string = pygame.image.tostring(img, 'RGBA', flip)
        size = img.get_size()
        filter_type = GL_NEAREST
        if filter: filter_type = GL_LINEAR
        self.id = glGenTextures(1)
        glBindTexture(GL_TEXTURE_2D, self.id)
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, filter_type)
        if mipmap:
            glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR_MIPMAP_NEAREST)
            gluBuild2DMipmaps(GL_TEXTURE_2D, channels, size[0], size[1], ptype, GL_UNSIGNED_BYTE, string)
        else:
            glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, filter_type)
            glTexImage2D(GL_TEXTURE_2D, 0, ptype, size[0], size[1], 0, ptype, GL_UNSIGNED_BYTE, string)
        
        #keep a ref to this, for cleaning up in the destructor.
        self.glDeleteTextures = glDeleteTextures
    
    def __del__(self):
        if hasattr(self, 'id'):
            self.glDeleteTextures(self.id)       


class FontTexture(object):            
    def __init__(self, font_filename, foreground=(1,1,1), background=None):
        foreground = tuple([int(i*255) for i in foreground])
        if background is not None:
            background = tuple([int(i*255) for i in background])
        
        font = pygame.font.Font(font_filename, 40)
        pack = qgl.texture.Pack(size=(1024,1024))
        
        alphabet = {}

        for c in 'abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ`1234567890-=+_)(*&^%$#@!~[]\\;\',./<>?:"{}| ':
            image = font.render(c, True, foreground).convert(32, pygame.SRCALPHA)
            image = pygame.transform.flip(image, False, True)
            alphabet[c] = pack.pack(image)
            
        self.alphabet = alphabet
        self.texture = GLTexture(pack.image, True,True)
        
        
        
class Compiler(Visitor):
    """
    Traverse the graph, compiling and leaf nodes so they can be rendered by 
    OpenGL. Sets all nodes to enabled, and all switches to default.
    """
    def __init__(self):
        self.stack = []
    def push_state(self, node):
        for child in node.branches:
            child.enable()
        self.stack.append(node)
        
    def pop_state(self):
        node = self.stack.pop(-1)
        if isinstance(node, qgl.scene.Switch):
            node.switch('default')
        
    def visit_Switch(self, node):
        branches = []
        for choice in node.choices.values():
            branches.extend(choice)
        node.branches = branches
    
    def visit_Group(self, node):
        node.compiler = self
        for leaf in node.leaves:
            if isinstance(leaf, qgl.scene.Texture):
                self.build_texture(leaf)
            if isinstance(leaf, qgl.scene.Sequence):
                self.build_animation(leaf)
            elif isinstance(leaf, qgl.scene.Text):
                self.build_text(leaf)
                
    def visit_Viewport(self, node):
        node.window_size = pygame.display.get_surface().get_size()
                
    def build_texture(self, node):
        key = (node.filename, node.mipmap, node.filter)
        if key in qgl_cache:
            node.texture = qgl_cache[key]
        else:
            texture = GLTexture.from_file(*key)
            qgl_cache[key] = texture
            node.texture = texture
            
    def build_animation(self, node):
        key = node.filename
        if key in qgl_cache:
            node.texture = qgl_cache[key]
        else:
            f = zipfile.ZipFile(node.filename,"r")
            fo = f.read("texture.rgba")
            size  = eval(f.read("size.repr"))
            image = pygame.image.fromstring(fo, size, "RGBA")
            texture = GLTexture(image, True, True)
            texture.coords = eval(f.read("coords.repr"))
            qgl_cache[key] = texture
            node.texture = texture
            
    
    def build_text(self, node):
        key = node.font, node.foreground, node.background
        if key in qgl_cache:
            node.font_object = qgl_cache[key]
        else:
            font = FontTexture(*key)
            qgl_cache[key] = font
            node.font_object = font
        
        lst = glGenLists(1)
        if lst == 0:
            raise RuntimeError("OpenGL will not allocate display list.")
        glNewList(lst, GL_COMPILE)
        glEnable(GL_TEXTURE_2D)
        glBindTexture(GL_TEXTURE_2D, node.font_object.texture.id)
        glBegin(GL_QUADS)
        ox,oy = 0.0,0.0
        for c in node._text:
            tex = node.font_object.alphabet[c]
            h = tex[2][1] - tex[0][1]
            w = tex[2][0] - tex[0][0]
            h *= node.size
            w *= node.size
            glTexCoord2f(*tex[0])
            glVertex2f(ox+0, oy+0)
            glTexCoord2f(*tex[1])
            glVertex2f(ox+0, oy+h)
            glTexCoord2f(*tex[2])
            glVertex2f(ox+w, oy+h)
            glTexCoord2f(*tex[3])
            glVertex2f(ox+w, oy)
            ox += w
        glEnd()
        glEndList()
        node.list = lst
            
        
class Render(Visitor):
    """
    Traverse the graph and render using OpenGL.
    """
    def push_state(self, node):
        glPushMatrix()
        glPushAttrib(GL_CURRENT_BIT|GL_TEXTURE_BIT|GL_ENABLE_BIT|GL_LIGHTING_BIT|GL_POLYGON_BIT)

    def pop_state(self):
        glPopAttrib()
        glPopMatrix()
    
    def visit_Root(self, node):
        glEnable(GL_DEPTH_TEST)
        glDepthFunc(GL_LEQUAL)
        glClearDepth(1.0)
        glFrontFace(GL_CW)
        glShadeModel(GL_SMOOTH)
        glCullFace(GL_BACK)
        glClearColor(*node.background_color)
        glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT)
        glEnable(GL_BLEND)
        glBlendFunc(GL_SRC_ALPHA,GL_ONE_MINUS_SRC_ALPHA)
        
    def visit_Transform(self, node):
        glTranslatef(*node.translate)
        glRotatef(node.angle, *node.axis)
        glScalef(*node.scale)
        
    def visit_Group(self, node):
        glTranslatef(*node.translate)
        glRotatef(node.angle, *node.axis)
        glScalef(*node.scale)
        for leaf in node.leaves:
            if leaf.__class__ is Texture:
                glEnable(GL_TEXTURE_2D)
                glBindTexture(GL_TEXTURE_2D, leaf.texture.id)
            elif leaf.__class__ is Color:
                glColor4f(leaf.rgba[0], leaf.rgba[1], leaf.rgba[2], leaf.rgba[3])
            elif leaf.__class__ is Polyline:
                glBegin(GL_LINE_STRIP)
                for v in leaf.vertices:
                    glVertex2f(v[0],v[1])
                glEnd()
            elif leaf.__class__ is Quad:
                v = leaf.vertices
                t = leaf.texture_coords
                glBegin(GL_QUADS)
                glTexCoord2f(t[0][0],t[0][1])
                glVertex2f(v[0][0],v[0][1])
                glTexCoord2f(t[1][0],t[1][1])
                glVertex2f(v[1][0],v[1][1])
                glTexCoord2f(t[2][0],t[2][1])
                glVertex2f(v[2][0],v[2][1])
                glTexCoord2f(t[3][0],t[3][1])
                glVertex2f(v[3][0],v[3][1])
                glEnd()
            elif leaf.__class__ is Sequence:
                v = leaf.vertices
                t = leaf.texture.coords[leaf.frame]
                glEnable(GL_TEXTURE_2D)
                glBindTexture(GL_TEXTURE_2D, leaf.texture.id)
                glBegin(GL_QUADS)
                glTexCoord2f(t[0][0], t[0][1])
                glVertex2f(v[0][0],v[0][1])
                glTexCoord2f(t[1][0],t[1][1])
                glVertex2f(v[1][0],v[1][1])
                glTexCoord2f(t[2][0],t[2][1])
                glVertex2f(v[2][0],v[2][1])
                glTexCoord2f(t[3][0],t[3][1])
                glVertex2f(v[3][0],v[3][1])
                glEnd()
            elif leaf.__class__ is Text:
                glCallList(leaf.list)
                
    def visit_Viewport(self, node):
        glViewport(*node.screen_dimensions)
        glMatrixMode(GL_PROJECTION)
        glLoadIdentity()
        gluPerspective(45.0, node.aspect, 0.1, 100)
        glMatrixMode(GL_MODELVIEW)
        glLoadIdentity()
        
        
class Picker(Render):
    """
    Traverse the graph and create a list (self.hits) of all items rendered at
    the area defined by self.position and self.area.
    """
    def __init__(self, *args, **kw):
        Render.__init__(self, *args, **kw)
        self.set_position((0,0), area=(1,1))
    
    def set_position(self, pos, area=(1,1)):
        self.position = pos
        self.area = area 
        self.guid = 0
        self.stack = []
        self.hits = []
        self.selectable_nodes = {}
        
    def push_state(self, node):
        self.guid += 1
        Render.push_state(self, node)
        self.stack.append(node)
        if isinstance(node, qgl.scene.Root):
            glInitNames()
            glSelectBuffer(256)
            glRenderMode(GL_SELECT)
        if node.selectable:
            node_id = self.guid
            self.selectable_nodes[node_id] = node
            glPushName(node_id)
        
    def pop_state(self):
        Render.pop_state(self)
        node = self.stack.pop(-1)
        if isinstance(node, qgl.scene.Root):
            glFlush()
            items = []
            hits = list(glRenderMode(GL_RENDER))
            for hit in hits:
                items.extend([self.selectable_nodes[i] for i in hit[2]])
            self.hits = items
        if node.selectable:
            glPopName()
            
    def visit_Viewport(self, node):
        glViewport(*node.screen_dimensions)
        glMatrixMode(GL_PROJECTION)
        glLoadIdentity()
        viewport = glGetIntegerv(GL_VIEWPORT)
        gluPickMatrix(self.position[0],node.window_size[1]-self.position[1],self.area[0],self.area[1],viewport)
        gluPerspective(45, node.aspect, 0.1, 100)
        glMatrixMode(GL_MODELVIEW)
        glLoadIdentity()
            


    
