import weakref
import random 
import Numeric
from qgl import texture
from copy import copy


class Node(object): 
    def __init__(self, *args):
        self.branches = []
        self.selectable = False
        self._accept = self.accept
        self.node_type = self.__class__.__name__
        self.parent = None
        self.add(*args)

    def clone(self):
        return copy(self)
    
    def __getitem__(self, i):
        return self.branches[i]
    
    def disable(self):
        if self.accept != self.disabled_accept:
            self._accept = self.accept
            self.accept = self.disabled_accept
    
    def enable(self):
        if self.accept != self._accept:
            self.accept = self._accept
    
    def disabled_accept(self, visitor):
        pass
        
    def accept(self, visitor):
        visitor.push_state(self)
        getattr(visitor, 'visit_' + self.node_type)(self)
        for node in self.branches:
            node.accept(visitor)
        visitor.pop_state()
        
    def add(self, *args):
        for arg in args:
            arg.parent = weakref.proxy(self)
            self.branches.append(arg)
        
    def remove(self, *args):
        for arg in args:
            self.branches.remove(arg)
    
    def update(self, **kw):
        self.__dict__.update(kw)
        

class Leaf(object):
    pass


class Root(Node):
    """
    The root node. There must only be one.
    """
    def __init__(self, *args):
        Node.__init__(self, *args)
        self.background_color = (0.0,0.0,0.0,1.0)
    

class Viewport(Node):
    """
    The Viewport specifies an area of the screen which all children are 
    rendered to.
    """
    def __init__(self, *args):
        Node.__init__(self, *args)
        self.screen_dimensions = (0, 0, 800, 600)
        self.aspect = 800.0 / 600.0
        

class Transform(Node):
    """
    A Transform node is used to scale, move and rotate its child nodes.
    """
    def __init__(self, *args):
        Node.__init__(self, *args)
        self.scale = (1.0,1.0,1.0)
        self.angle = 0.0
        self.axis = (0.0,0.0,1.0)
        self.translate = (0.0,0.0,0.0)

    
class Group(Transform):
    """
    A Group node is the node which contains renderable items.
    """
    def __init__(self, *args):
        Transform.__init__(self, *args)
        self.leaves = []
        
    def add(self, *args):
        for leaf in args:
            if isinstance(leaf, Leaf):
                leaf.parent = weakref.proxy(self)
                self.leaves.append(leaf)
            else:
                Node.add(self, leaf)
        
    def remove(self, *args):
        for leaf in args:
            if isinstance(leaf, Leaf):
                self.leaves.remove(leaf)
            else:
                Node.remove(self, leaf)
        

class Switch(Node):
    """
    A Switch node has named groups of children and can change the traversal 
    path depending on the active name. The initial name is 'default'.
    To add children to a node, switch to the name, then add children as 
    normal.
    """
    def __init__(self, *args):
        Node.__init__(self, *args)
        self.choices = dict(default=list(args))
        self.branches = self.choices['default']
    
    def switch(self, name):
        self.branches = self.choices.setdefault(name, [])
        

class World(Node):
    """
    The World node is a helper node which contains pre configured planes for
    sky, ground, background and foreground.
    """
    def __init__(self):
        Node.__init__(self)
        self.node_type = 'Node'
        self.sky = sky = Group()
        sky.update(angle=90.0, axis=(1,0,0), translate=(0.0,2.0,0.0))
        self.ground = ground = Group()
        ground.update(angle=-90.0, axis=(1,0,0), translate=(0.0,-2.0,0.0))
        self.background = background = Group()
        background.update(translate=(0.0,0.0,-40.0))
        self.foreground = foreground = Group()
        foreground.update(translate=(0.0,0.0,-10.0))
        self.camera = Transform(sky, ground, background, foreground)
        self.add(self.camera)


class Texture(Leaf):
    """
    A Texture leaf is added to a group node to change the texture which is 
    used to render subsequent quads.
    """
    def __init__(self, filename, filter=True, mipmap=True):
        self.filename = filename
        self.filter = filter
        self.mipmap = mipmap


class Color(Leaf):
    """
    The Color leaf sets the color (including alpha value) for a nodes 
    children and leaves.
    """
    def __init__(self, rgba=(1.0,1.0,1.0,1.0)):
        self.rgba = rgba


class Quad(Leaf):
    """
    The Quad leaf draws a rectangle using the current texture and color. It 
    is drawn from its centre point.
    """
    def __init__(self, (w, h)):
        x,y = w*0.5,h*0.5
        self.vertices = (-x, -y), (-x, y), (x, y), (x, -y)
        self.texture_coords = texture.Coords()


class Polyline(Leaf):
    """
    The Polyline leaf draws a series of connected lines.
    """
    def __init__(self, vertices, width=1.0):
        self.width = 1.0
        self.vertices = vertices 
        

class Sequence(Leaf):
    """
    The sequence leaf draws an image sequence loaded from a file created
    using the make_animation tool.
    """
    def __init__(self, (w,h), filename):
        x,y = w*0.5,h*0.5
        self.vertices = (-x, -y), (-x, y), (x, y), (x, -y)
        self.filename = filename
        self.frame = 0
        
    
class Text(Leaf):
    """
    The Text leaf draws text using a specified font from the bottom left 
    corner.
    """
    def __init__(self, text, font, foreground=(1,1,1), background=None, size=32):
        self._text = text
        self.foreground = foreground
        self.background = background
        self.size = size
        self.font = font
    
    def set_text(self, text):
        self._text = text
        self.parent.accept(self.parent.compiler)
    def get_text(self):
        return self._text
    text = property(fset=set_text, fget=get_text)
        
